"""
Main class for network inference and simulation
"""
import numpy as np
from ..inference import inference_optim, infer_kinetics, log_gamma_poisson_pdf

np.set_printoptions(formatter={'float': lambda x: "{0:0.3f}".format(x)})

class NetworkModel:
    """
    Handle networks within the package.
    """
    def __init__(self, n_genes=None, times=None):
        # Kinetic parameters
        self.d = None
        # Mixture parameters
        self.data_bool = None
        # Optim parameters
        self.n_train = None
        # Network parameters
        self.mask = None
        self.a = None
        self.basal = None
        self.inter = None
        self.basal_t = None
        self.inter_t = None
        self.variations = None
        # Default behaviour
        if n_genes is not None:
            G = n_genes + 1 # Genes plus stimulus
            # Default degradation rates
            self.d = np.zeros((2,G))
            self.d[0] = np.log(2)/9 # mRNA degradation rates
            self.d[1] = np.log(2)/46 # protein degradation rates
            # Default network parameters
            self.basal = np.zeros(G)
            self.inter = np.zeros((G,G))
            self.variations = np.zeros((G, G))


    def core_basins_binary(self, data, alph, b, g):
        """
        Compute the basal parameters of filtered genes.
        """
        for n, c in enumerate(data):
            for z, k in enumerate(alph):
                self.weight[n, g, z] = log_gamma_poisson_pdf(c, k, b)
            self.data_bool[n, g] = alph[np.argmax(self.weight[n, g, :])]


    def get_basins_binary(self, data, verb=False):
        """
        Compute the basal parameters of filtered genes.
        """
        seuil = 1e-3
        C, G = data.shape
        times = list(set(data[:, 0]))
        times.sort()
        vect_t = data[:, 0]

        self.data_bool = np.ones_like(data, dtype='float')
        self.data_bool[vect_t == 0, 0] = 0
        self.weight = np.zeros((C, G, 2), dtype='float')
        a = np.ones((3, G))
        for g in range(1, G):
            x = data[:, g]
            at, a[-1, g] = infer_kinetics(x, vect_t, verb=verb)
            a[0, g] = max(np.min(at), seuil)
            a[1, g] = max(np.max(at), seuil)
            if verb: print('Gene {} calibrated...'.format(g), a[:, g])
            self.core_basins_binary(x, a[:-1, g], a[-1, g], g)
        self.a = a


    def core_optim(self, x, vect_t, times, G_tot):
        """
        Fit the network model to the data.
        Return the list of successive objective function values.
        """

        # Initialization
        k_max = np.max(x, 0)
        ko = np.min(x, 0) / k_max
        y, vect_kon = x / k_max, x / k_max

        variations_tot = np.zeros((G_tot, G_tot))
        basal_tot = np.zeros(G_tot)
        inter_tot = np.zeros((G_tot, G_tot))
        basal_tot_t = np.zeros((len(times), G_tot))
        inter_tot_t = np.zeros((len(times), G_tot, G_tot))

        # Inference procedure
        basal, inter, variations_time, basal_t, inter_t = inference_optim(vect_t, times, y, vect_kon, ko)

        # Build the results
        cnt_i, cnt_j = 0, 0
        for i in range(0, G_tot):
            cnt_j = 0
            if self.mask[i]:
                for j in range(0, G_tot):
                    if self.mask[j]:
                        inter_tot[i, j] = inter[cnt_j, cnt_i]
                        inter_tot_t[:, i, j] = inter_t[:, cnt_j, cnt_i]
                        variations_tot[i, j] = variations_time[cnt_j, cnt_i]
                        cnt_j += 1
                basal_tot[i] = basal[cnt_i]
                basal_tot_t[:, i] = basal_t[:, cnt_i]
                cnt_i += 1

        return basal_tot, inter_tot, variations_tot, basal_tot_t, inter_tot_t


    def fit(self, data, verb=False):
        """
        Fit the network model to the data.
        Return the list of successive objective function values.
        """
        C, G_tot = data.shape

        # Get kinetic parameters
        self.get_basins_binary(data, verb=verb)

        self.mask = np.ones(G_tot, dtype='bool')
        vect_t = data[:, 0]
        times = list(set(vect_t))
        times.sort()
        for g in range(1, G_tot):
            meang = [np.mean(data[vect_t == time, g]) for time in times]
            if np.max(meang) - np.min(meang) < .1:
                self.mask[g] = 0

        if verb: print('number genes of interest', np.sum(self.mask))
        self.d = self.d[:, self.mask]
        data_bool = self.data_bool[:, self.mask]

        self.basal, self.inter, self.variations, self.basal_t, self.inter_t = \
            self.core_optim(data_bool, data[:, 0], times, G_tot)

        if verb: print('TOT', self.inter.T, self.basal)




"""
Core functions for network inference.

Considered sizes
----------------
    C : number of cells
    G : number of genes including stimulus

Parameters
----------
    x : array (C, G)
        Observed mRNA levels (column 0 = time points).
    inter : dictionnary of arrays (G, G)
        Given key t, inter[t][i,j] denotes interaction i -> j at time t.
    basal : array (G,)
        Basal activity for each gene.
"""
import numpy as np
from scipy.optimize import minimize
from scipy.special import expit
from numba import njit

sl = 5e-3 # pseudo l1 coefficient
p = .4

@njit
def penalization_l1(x, s):
    return (x-s/2)*(x>s) - (x+s/2)*(-x>s) + ((x**2)/(2*s))*(x<=s and -x<=s)

@njit
def grad_penalization_l1(x, s):
    return 1*(x>s) - 1*(-x>s) + (x/s)*(x<=s and -x<=s)

@njit
def penalization(Q, X, X_init, time_init, l, sc, G, cnt_move, j):

    l_inter, l_diag = l[0], l[1]
    if not time_init: Q += penalization_l1(X[-1] - X_init[j, -1], sc) * l_inter * p
    if time_init == 1: Q += penalization_l1(X[0] - X_init[j, 0], sc) * l_inter * p
    tmp_list = list(range(1, j)) + list(range(j + 1, G))
    for i in tmp_list:
        tmp = cnt_move[i] * (1 + abs(X_init[i, j]) / (1 + abs(X_init[j, i])))
        Q += penalization_l1(X[i] - X_init[j, i], sc) * l_inter * tmp
    tmp_diag = cnt_move[j] / (1 + np.sum(np.abs(X_init[j, :-1])) - abs(X_init[j, j]))
    Q += l_diag * (penalization_l1(X[j] - X_init[j, j], sc) + (X[j] - X_init[j, j]) ** 2) * tmp_diag
    return Q

@njit
def grad_penalization(dq, X, X_init, time_init, l, sc, G, cnt_move, j):

    l_inter, l_diag = l[0], l[1]
    if not time_init: dq[-1] += grad_penalization_l1(X[-1] - X_init[j, -1], sc) * l_inter * p
    else: dq[-1] = 0
    if time_init == 1: dq[0] += grad_penalization_l1(X[0] - X_init[j, 0], sc) * l_inter * p
    else:  dq[0] = 0
    tmp_list = list(range(1, j)) + list(range(j + 1, G))
    for i in tmp_list:
        tmp = cnt_move[i] * (1 + abs(X_init[i, j]) / (1 + abs(X_init[j, i])))
        dq[i] += grad_penalization_l1(X[i] - X_init[j, i], sc) * l_inter * tmp
    tmp_diag = cnt_move[j] / (1 + np.sum(np.abs(X_init[j, :-1])) - abs(X_init[j, j]))
    dq[j] += l_diag * (grad_penalization_l1(X[j] - X_init[j, j], sc) + 2 * (X[j] - X_init[j, j])) * tmp_diag
    return dq

def objective(X, y, vect_kon, ko, X_init, time_init, l, sc, G, cnt_move, j):
    """
    Objective function to be maximized (all cells, one gene, all timepoints).
    """
    sigma = expit(X[-1] + y @ X[:-1])
    Q = np.sum((ko + (1 - ko) * sigma - vect_kon) ** 2)
    return penalization(Q, X, X_init, time_init, l, sc, G, cnt_move, j)


def grad_theta(X, y, vect_kon, ko, X_init, time_init, l, sc, G, cnt_move, j):
    """
    Objective gradient for gene i for all cells.
    """

    dq = np.zeros(G + 1)
    sigma = expit(X[-1] + y @ X[:-1])
    tmp = 2 * (ko + (1 - ko) * sigma - vect_kon) * (1 - ko) * sigma * (1 - sigma)
    for i in range(0, G):
        dq[i] += np.sum(y[:, i] * tmp)
    dq[-1] += np.sum(tmp)

    return grad_penalization(dq, X, X_init, time_init, l, sc, G, cnt_move, j)


def build_cnt(cnt, cnt_move, vect_kon, vect_t, times, G):

    for j in range(1, G):
        for i in range(1, G):
            tmp_cnt = 0
            if cnt > 2: # -2 for letting the first timepoints after the stimulus unchanged
                for t in range(0, cnt - 2):
                    tmp = np.sum(vect_t == times[t])
                    tmp1 = np.sum(vect_kon[vect_t == times[t], i] == 1) / tmp
                    tmp2 = np.sum(vect_kon[vect_t == times[t], j] == 1) / tmp
                    if tmp1 > p:
                        tmp_cnt += tmp1 * tmp2
                tmp_cnt /= cnt-1
            cnt_move[j, i] = (1 + tmp_cnt) * max(p, min(1 +
                np.sum(vect_kon[vect_t == times[cnt - 2], i] == 1) / np.sum(vect_t == times[cnt - 2])
              - np.sum(vect_kon[vect_t == times[cnt - 1], i] == 1) / np.sum(vect_t == times[cnt - 1]), 1))
    return cnt_move


def core_inference(variations, times, vect_t, y, vect_kon, ko, Xo, l, sl, G, cnt_init, cnt_end):

    cnt_move = np.ones((G, G))
    theta_t = np.zeros((cnt_end - cnt_init, G, G + 1))
    for cnt, time in enumerate(times[cnt_init:cnt_end]):
        p = l * np.sum(vect_t == time) * 2.5 / 100
        if cnt + cnt_init > 1:
            cnt_move = build_cnt(cnt + cnt_init, cnt_move, vect_kon, vect_t, times, G)

        X_init = Xo.copy()
        for j in range(1, G):
            res = minimize(objective, X_init[j, :],
                               args=(y[vect_t == time, :],
                                     vect_kon[vect_t == time, j],
                                     ko[j], X_init, cnt + cnt_init, p, sl, G, cnt_move[j], j),
                               jac=grad_theta,
                               method='L-BFGS-B',
                               tol=1e-5)

            if not res.success: print('Warning, minimization time {} failed'.format(time))
            variations[cnt][j, :] += np.abs(res.x[:-1] - Xo[j, :-1])
            Xo[j, :] = res.x[:]
            theta_t[cnt, j, :] = Xo[j, :]

    return Xo, variations, theta_t


def inference_optim(vect_t, times, y, vect_kon, ko):
    """
    Network inference procedure.
    Return the inferred network (basal + network) and the time at which each edge has been detected with strongest intensity.
    """
    G = np.size(y, 1)
    inter = np.zeros((G, G))
    basal = np.zeros(G)
    inter_t = np.zeros((len(times), G, G))
    basal_t = np.zeros((len(times), G))
    variations = np.zeros((len(times), G, G))
    time_variations = np.zeros((G, G))

    # penalites
    l = np.array([1, np.sqrt(G - 1)])  # l_inter, l_basal, l_diag

    Xo = np.zeros((G, G + 1))
    Xo, variations, theta_t = core_inference(variations, times, vect_t, y, vect_kon, ko, Xo, l, sl, G, 0, len(times))

    for i in range(0, G):
        for j in range(0, G):
            time_variations[i, j] = np.argmax(variations[1:, i, j])

    inter[:, :] = Xo[:, :-1]
    basal[:] = Xo[:, -1]
    inter_t[:, :, :] = theta_t[:, :, :-1]
    basal_t[:, :] = theta_t[:, :, -1]

    return basal, inter, time_variations, basal_t, inter_t


"""
Inference of the network model.
"""
from .network import inference_optim
from .kinetics import infer_kinetics, log_gamma_poisson_pdf


__all__ = ['inference_optim', 'infer_kinetics', 'log_gamma_poisson_pdf']



"""
CARDAMOM - Benchmark paper version
========

Gene regulatory network inference from single-cell data
-------------------------------------------------------

Author: Elias Ventre
"""
__version__ = '1.0 - benchmark paper'

from .model import NetworkModel

__all__ = ['NetworkModel']


"""
Inference of basal parameters
This file comes from the 'Harissa' package written by Ulysse Herbach
        https://github.com/ulysseherbach/harissa
"""

import numpy as np
from numpy import log
from scipy.special import psi, polygamma, gammaln


def log_gamma_poisson_pdf(k, a, b):

    ga, gk = gammaln(a), gammaln(k + 1)
    return gammaln(a + k) - (ga + gk) + a * np.log(b) - (a + k) * np.log(b + 1)


def estim_gamma_poisson(x):
    """
    Estimate parameters a and b of the Gamma-Poisson(a,b) distribution,
    a.k.a. negative binomial distribution, using the method of moments.
    """
    m1 = np.mean(x)
    m2 = np.mean(x*(x-1))
    if m1 == 0: return 0, 1
    r = m2 - m1**2
    if r > 0: a = m1 ** 2 / r
    else:
        v = np.var(x)
        if v == 0: return 0, 1
        a = m1 ** 2 /v
    b = a / m1
    return a, b


def infer_kinetics(x, times, tol=1e-5, max_iter=1000, verb=False):
    """
    Infer parameters a[0], ..., a[m-1] and b of a Gamma-Poisson model
    with time-dependant a and constant b for a given gene at m time points.

    Parameters
    ----------
    x[k] = gene expression in cell k
    times[k] = time point of cell k
    """

    t = np.sort(list(set(times)))
    m = t.size
    n = np.zeros(m) # Number of cells for each time point
    a = np.zeros(m)
    b = np.zeros(m)
    # Initialization of a and b
    for i in range(m):
        cells = (times == t[i])
        n[i] = np.sum(cells)
        a[i], b[i] = estim_gamma_poisson(x[cells])
    b = np.mean(b)
    # Newton-like method
    k, c = 0, 0
    sx = np.sum(x)
    while (k == 0) or (k < max_iter and c > tol):
        da = np.zeros(m)
        for i in range(m):
            if a[i] > 0:
                cells = (times == t[i])
                z = a[i] + x[cells]
                p0 = np.sum(psi(z))
                p1 = np.sum(polygamma(1, z))
                d = n[i]*(log(b)-log(b+1)-psi(a[i])) + p0
                h = p1 - n[i]*polygamma(1, a[i])
                da[i] = -d/h
        anew = a + da
        if np.sum(anew < 0) == 0: a[:] = anew
        else:
            max_test = 5
            test = 0
            da *= 0.5
            while (np.sum(a + da < 0) > 0) and (test < max_test):
                da *= 0.5
                test += 1
            if test < max_test: a[:] = a + da
            else: print('Warning: parameter a not improved')
        if np.sum(a == 0) == 0:
            b = np.sum(n*a)/sx
        else: b = b
        c = np.max(np.abs(da))
        k += 1
    if (k == max_iter) and (c > tol):
        print('Warning: bad convergence (b = {})'.format(b))
        a, b = a/b, 1
    # if verb: print('Estimation done in {} iterations'.format(k))
    if np.sum(a < 0) > 0: print('WARNING: a < 0')
    if b < 0: print('WARNING: b < 0')
    if np.all(a == 0): print('WARNING: a == 0')
    # if k > 20 and np.max(a/b) > 2: print(k, np.max(a/b))

    return a, b
